/*
 * avrgal - based on avrdude
 * avrdude - A Downloader/Uploader for AVR device programmers
 * Copyright (C) 2000-2004  Brian S. Dean <bsd@bsdhome.com>
 * Copyright (C) 2010  David Anders <danders@rad-tech-labs.com>
 * Copyright (C) 2010  TinCanTools <sales@tincantools.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <termios.h>
#include <sys/ioctl.h>

#include "protocol.h"
#include "serial.h"

static char *progname = "PROTOCOL";

static int protocol_is_page_empty(unsigned int address, int page_size, 
                                const unsigned char *buf)
{
    int i;
    for(i = 0; i < page_size; i++) {
        if(buf[address + i] != 0xFF) {
            /* Page is not empty. */
            return(0);
        }
    }
    
    /* Page is empty. */
    return(1);
}

int protocol_open(char * port, long baud)
{
    return ser_open(port,baud);
}

void protocol_close(int fd)
{
    ser_close(fd);
}


int protocol_getsync(int fd)
{
  unsigned char buf[32], resp[32];

  /*
   * get in sync */
  buf[0] = Cmnd_STK_GET_SYNC;
  buf[1] = Sync_CRC_EOP;
  
  /*
   * First send and drain a few times to get rid of line noise 
   */
   
  ser_send(fd, buf, 2);
  ser_drain(fd, 0);
  ser_send(fd, buf, 2);
  ser_drain(fd, 0);

  ser_send(fd, buf, 2);

  if (ser_recv(fd, resp, 1) < 0)
    return -1;
  if (resp[0] != Resp_STK_INSYNC) {
    ser_drain(fd, 0);
    return -1;
  }

  if (ser_recv(fd, resp, 1) < 0)
    return -1;
  if (resp[0] != Resp_STK_OK) {
    return -1;
  }

  return 0;
}



int protocol_loadaddr(int fd, unsigned int addr)
{
  unsigned char buf[16];
  int tries;

  tries = 0;
 retry:
  tries++;
  buf[0] = Cmnd_STK_LOAD_ADDRESS;
  buf[1] = addr & 0xff;
  buf[2] = (addr >> 8) & 0xff;
  buf[3] = Sync_CRC_EOP;

  ser_send(fd, buf, 4);
    
  if (ser_recv(fd, buf, 1) < 0)
    exit(1);
  if (buf[0] == Resp_STK_NOSYNC) {
    if (tries > 33) {
      fprintf(stderr, "%s: protocol_loadaddr(): can't get into sync\n",
              progname);
      return -1;
    }
    if (protocol_getsync(fd) < 0)
      return -1;
    goto retry;
  }
  else if (buf[0] != Resp_STK_INSYNC) {
    fprintf(stderr,
            "%s: protocol_loadaddr(): (a) protocol error, "
            "expect=0x%02x, resp=0x%02x\n", 
            progname, Resp_STK_INSYNC, buf[0]);
    return -1;
  }

  if (ser_recv(fd, buf, 1) < 0)
    exit(1);
  if (buf[0] == Resp_STK_OK) {
    return 0;
  }

  fprintf(stderr,
          "%s: loadaddr(): (b) protocol error, "
          "expect=0x%02x, resp=0x%02x\n", 
          progname, Resp_STK_INSYNC, buf[0]);

  return -1;
}


int protocol_paged_write(int fd,unsigned char * codebuf , 
                              int page_size, int n_bytes)
{
  unsigned char buf[page_size + 16];
  int memtype;
  unsigned int addr;
  int a_div;
  int block_size;
  int tries;
  unsigned int n;
  unsigned int i;
  int flash;

  if (page_size == 0) {
      page_size = 128;
  }

    memtype = 'F';
    flash = 1;
    a_div = 2;

  if (n_bytes > 32768) {
    n_bytes = 32768;
    n = 32768;
  }
  else {
    if ((n_bytes % page_size) != 0) {
      n = n_bytes + page_size - (n_bytes % page_size);
    }
    else {
      n = n_bytes;
    }
  }
#if DEBUG
  fprintf(stderr, 
          "n_bytes   = %d\n"
          "n         = %u\n"
          "a_div     = %d\n"
          "page_size = %d\n",
          n_bytes, n, a_div, page_size);
#endif
  for (addr = 0; addr < n; addr += page_size) {
//    report_progress (addr, n_bytes, NULL);
    
    if ((addr + page_size > n_bytes)) {
	   block_size = n_bytes % page_size;
	}
	else {
	   block_size = page_size;
	}
  
    /* Only skip on empty page if programming flash. */
    if (flash) {
      if (protocol_is_page_empty(addr, block_size, codebuf)) {
          continue;
      }
    }
    tries = 0;
  retry:
    tries++;
    protocol_loadaddr(fd, addr/a_div);

    /* build command block and avoid multiple send commands as it leads to a crash
        of the silabs usb serial driver on mac os x */
    i = 0;
    buf[i++] = Cmnd_STK_PROG_PAGE;
    buf[i++] = (block_size >> 8) & 0xff;
    buf[i++] = block_size & 0xff;
    buf[i++] = memtype;
    memcpy(&buf[i], &codebuf[addr], block_size);
    i += block_size;
    buf[i++] = Sync_CRC_EOP;
    ser_send( fd, buf, i);

    if (ser_recv(fd, buf, 1) < 0)
      exit(1);
    if (buf[0] == Resp_STK_NOSYNC) {
      if (tries > 33) {
        fprintf(stderr, "\n%s: protocol_paged_write(): can't get into sync\n",
                progname);
        return -3;
      }
      if (protocol_getsync(fd) < 0)
	return -1;
      goto retry;
    }
    else if (buf[0] != Resp_STK_INSYNC) {
      fprintf(stderr,
              "\n%s: protocol_paged_write(): (a) protocol error, "
              "expect=0x%02x, resp=0x%02x\n", 
              progname, Resp_STK_INSYNC, buf[0]);
      return -4;
    }
    
    if (ser_recv(fd, buf, 1) < 0)
      exit(1);
    if (buf[0] != Resp_STK_OK) {
      fprintf(stderr,
              "\n%s: protocol_paged_write(): (a) protocol error, "
              "expect=0x%02x, resp=0x%02x\n", 
              progname, Resp_STK_INSYNC, buf[0]);
      return -5;
    }
  }

  return n_bytes;
}
